# import os
from pathlib import Path
# import textwrap
import click

from . import seeqret_init
from .fileutils import is_writable
from .run_utils import current_user


@click.command()
@click.pass_context
@click.option('--email', prompt=True, help='Your email address')
@click.option('--pubkey', prompt=True, help='Your public key')
def init(ctx, email, pubkey):
    """Initialize a server vault
    """
    dirname = Path('/srv')
    vault_dir = dirname / '.seeqret'

    # we want to create dirname / seeqret

    if not dirname.exists():
        # click.echo(f'The parent of the vault: {dirname} must exist.')
        ctx.fail(f'The parent of the vault: {dirname} must exist.')
        # return

    # if not is_writable(dirname):
    #     click.echo(f'The parent of the vault: {dirname} is not writable.')
    #     if click.confirm("Do you want me to try to fix this?"):
    #         pass
    #     else:
    #         ctx.fail(f'The parent of the vault: {dirname} must be writable.')
    #     # ctx.fail(f'The parent of the vault: {dirname} must be writable.')
    #     # return

    if vault_dir.exists():
        if not is_writable(vault_dir):
            ctx.fail(
                f'The vault: {vault_dir} exists and is not writeable, '
                'you must delete it manually.'
            )
            # return
        click.confirm(
            f'The vault: {vault_dir} already exists, overwrite contents?',
            default=True, abort=True)
        # remove_directory(vault_dir)

    curuser = current_user()
    # pkey_fname = os.path.expanduser('~/.ssh/seeqret-private.key')
    # pubkey_fname = os.path.expanduser('~/.ssh/seeqret-public.key')

    # if not os.path.exists(pkey_fname):
    #     click.secho(f"Private key {pkey_fname} does not exist", fg='red')
    #     click.echo(textwrap.dedent("""
    #         Please create a private key and public key in
    #
    #             ~/.ssh/seeqret-private.key
    #
    #         and
    #             ~/.ssh/seeqret-public.key
    #
    #         (run `seeqret keys` locally to display your keys).
    #     """))
    #     return
    # if not os.path.exists(pubkey_fname):
    #     click.secho(f"Public key {pubkey_fname} does not exist", fg='red')
    #     return

    # user_pkey = load_private_key(pkey_fname)
    # user_pubkey = load_public_key(pubkey_fname)
    #
    seeqret_init.secrets_server_init(dirname, vault_dir, curuser, email, pubkey)
    #
    #
    # seeqret_init.secrets_init(dirname, user, email, pubkey, key)



from click.testing import CliRunner

from seeqret.db_utils import debug_fetch_users
from seeqret.main import cli, user, users, init
from tests.clirunner_utils import print_result
from seeqret.run_utils import current_user


def test_add_user():
    runner = CliRunner(env=dict(TESTING="TRUE"))
    with runner.isolated_filesystem():
        result = runner.invoke(init, [
            '.',
            '--user=' + current_user(),
            '--email=test@example.com',
        ])
        if result.exit_code != 0:  print_result(result)
        assert result.exit_code == 0

        result = runner.invoke(users)
        if result.exit_code != 0:  print_result(result)
        assert result.exit_code == 0
        assert 'test@example.com' in result.output

        # print("DBUSERS:", debug_fetch_users())
        assert len(debug_fetch_users()) == 1
        result = runner.invoke(user,[
            '--username=tkbe',
            '--email=bjorn@tkbe.org',
            '--pubkey=ilxSnX9+NrwmeIzOFtWrl0lPPkxTEATmC39BILX6rWk='
            # '--url=https://raw.githubusercontent.com/tkbeorg/tkbe/refs/heads/main/public.key'
        ])
        if result.exit_code != 0:  print_result(result)
        print("DBUSERS:", debug_fetch_users())
        #
        assert len(debug_fetch_users()) == 2
        # print(result.output)
        assert result.exit_code == 0

        assert 'bjorn@tkbe.org' in result.output
        # assert 'test@example.com' in result.output

        # assert len(result.output) == 42

        # result = runner.invoke(user, [
        #     '--url='
        # ])




# def _validate_import_file(indata):
#     """Validate json input.
#     """
#     errors = False
#
#     # don't abort early, we want the user to receive _all_ the errors.
#     if 'from' not in indata:
#         click.secho('Invalid file format, missing "from"', fg='red')
#         errors = True
#     if 'username' not in indata['from']:
#         click.secho('Invalid file format, missing "from.username"', fg='red')
#         errors = True
#     if 'email' not in indata['from']:
#         click.secho('Invalid file format, missing "from.email"', fg='red')
#         errors = True
#     if 'pubkey' not in indata['from']:
#         click.secho('Invalid file format, missing "from.pubkey"', fg='red')
#         errors = True
#     if 'data' not in indata:
#         click.secho('Invalid file format, missing "data"', fg='red')
#         errors = True
#     if 'signature' not in indata:
#         click.secho('Invalid file format, missing "signature"', fg='red')
#         errors = True
#     if 'to' not in indata:
#         click.secho('Invalid file format, missing: "to"', fg='red')
#         errors = True
#     if 'username' not in indata['to']:
#         click.secho('Invalid file format, missing "to.username"', fg='red')
#         errors = True
#
#     return not errors
#
#
# def ximport_secrets(indata):
#     """seeqret import-file <fname:json>
#     """
#     if not _validate_import_file(indata):
#         click.secho('Invalid file format.', fg='red')
#         return
#     click.secho("file format ok...", fg='green')
#
#     signature = indata['signature']
#     data = indata['data']
#     from_user = indata['from']
#     to_user = indata['to']['username']
#
#     cn = sqlite3.connect('seeqrets.db')
#     admin = fetch_admin(cn)
#     if to_user != admin['username']:
#         click.secho(
#             f'Incorrect to.username {to_user}, not {admin["username"]}',
#             fg='red'
#         )
#         return
#     click.secho("Correct `to` field.")
#
#     # cipher = load_symetric_key('seeqret.key')
#     sender_pubkey = public_key(from_user['pubkey'])
#     receiver_private_key = load_private_key('private.key')
#     for item in data:
#         item['val'] = asymetric_decrypt_string(item['val'],
#                                                receiver_private_key,
#                                                sender_pubkey)
#
#     click.secho("verifying signature...")
#     verify_hash(signature, indata)
#     click.secho('Successfully validated file', fg='green')
#
#     if not fetch_user(cn, from_user['username']):
#         add_user(
#             from_user['pubkey'],
#             from_user['username'],
#             from_user['email']
#         )
#     else:
#         click.secho('Found sender', fg='green')
#     cn.close()
#
#     # click.echo(json.dumps(indata, indent=4))
#
#     for secret in data:
#         add_key(
#             secret['key'],
#             secret['val'],
#             secret['app'],
#             secret['env'],
#             # secret['type']
#         )
#
#
# def verify_hash(hash: str, message):
#     msg = _extract_data(message)
#     h = hash_message(msg.encode('utf-8'))
#     if hash != h:
#         msg = f'Invaolid hash ({hash}) expected: {h}'
#         ctx = None
#         try:
#             ctx = click.get_current_context()
#         except RuntimeError:
#             pass
#         if ctx is not None:
#             ctx.fail(click.style(msg, fg='red'))
#         else:
#             raise RuntimeError(msg)
#     return True
#
#
# def _extract_data(message):
#     """Extract all message values (so they can be hashed)
#     """
#     data = []
#     for secret in message['data']:
#         data.append(f"{secret['app']}:{secret['env']}:{secret['key']}:{secret['val']}\n")  # noqa
#     data.sort()
#     sender = message['from']
#     data.append(
#         f'From:{sender["username"]}|{sender["email"]}|{sender["pubkey"]}\n'
#     )
#     data.append(f'To:{message["to"]["username"]}\n')
#     msg = ''.join(data)
#     return msg
#



from seeqret.storage.get_secret import get_secret

__version__ = '0.2.0'


def get(key, app='*', env='*'):
    """Get a value from the configuration.

    Important: to keep your secrets out of tracebacks etc., you should call
    this function at the last possible moment. I.e., instead of having a
    settings file with:

        DB_PASSWORD = seeqret.get("db_password")  # BAD!

    do:

        cn = mydb.connect(..., password=seeqret.get("db_password"))  # GOOD!

    Args:
        key: The key to get.
        app: The application to get the key from.
        env: The environment to get the key from.

    Returns:
        The value of the key.
    """
    return get_secret(key, app, env)



import os

from cryptography.fernet import Fernet

from seeqret.seeqrypt.aes_fernet import encrypt_string, decrypt_string
from seeqret.seeqrypt.utils import (
    _chunks,
    format_encrypted_data,
    generate_symetric_key,
    load_symetric_key,
    get_or_create_symetric_key,
    generate_symetric_key,
    # generate_asymetric_keys,
    # load_public_key,
    # load_private_key,
)


def test_chunks():
    assert list(_chunks([1, 2, 3, 4, 5], 2)) == [[1, 2], [3, 4], [5]]


def test_format_encrypted_data():
    fmt = format_encrypted_data(b'0123456789'*10).splitlines()
    assert len(fmt) == 2
    assert len(fmt[0]) == 64
    assert len(fmt[1]) == 100 % 64


def test_generate_symetric_key():
    key = generate_symetric_key('test_generate_symetric_key.key')
    assert isinstance(key, Fernet)
    assert os.path.exists('test_generate_symetric_key.key')
    os.unlink('test_generate_symetric_key.key')


def test_load_symetric_key():
    key = generate_symetric_key('test_load_symetric_key.key')
    k2 = load_symetric_key('test_load_symetric_key.key')
    msg = b'hello world'
    assert decrypt_string(k2, encrypt_string(key, msg)) == msg
    os.unlink('test_load_symetric_key.key')


def test_get_or_create_symetric_key():
    key = get_or_create_symetric_key('test_get_or_create_symetric_key.key')
    k2 = get_or_create_symetric_key('test_get_or_create_symetric_key.key')
    msg = b'hello world'
    assert decrypt_string(k2, encrypt_string(key, msg)) == msg
    os.unlink('test_get_or_create_symetric_key.key')
#
#
# def test_load_public_key():
#     _pub, priv = generate_asymetric_keys('test_load_public_key.key')
#     pub = load_public_key('test_load_public_key.key')
#
#
#
# def test_load_private_key():
#     assert load_private_key() == b'12345'



import re
import sys

from click.testing import CliRunner
from seeqret.main import users, init
from tests.clirunner_utils import print_result
from seeqret.run_utils import current_user


def test_users():
    runner = CliRunner(env=dict(TESTING="TRUE"))
    with runner.isolated_filesystem():
        result = runner.invoke(init, [
            '.',
            '--user=' + current_user(),
            '--email=test@example.com',
        ])
        if result.exit_code != 0:  print_result(result)
        assert result.exit_code == 0
    
        result = runner.invoke(users)
        if result.exit_code != 0:  print_result(result)
        assert result.exit_code == 0

        result = runner.invoke(users, ['--export'])
        if result.exit_code != 0:  print_result(result)
        assert result.exit_code == 0




def table_exists(cn, table_name):
    c = cn.cursor()
    c.execute(f"""
        select count(*)
        from sqlite_master
        where type="table" and name="{table_name}"
    """)
    return c.fetchone()[0] > 0


def column_exists(cn, table_name, column_name):
    c = cn.cursor()
    c.execute(f"""
        select count(*)
        from pragma_table_info("{table_name}")
        where name="{column_name}"
    """)
    return c.fetchone()[0] > 0


def index_exists(cn, table_name, index_name):
    c = cn.cursor()
    c.execute(f"""
        select count(*)
        from sqlite_master
        where type="index" and name="{index_name}"
    """)
    return c.fetchone()[0] > 0


def current_version(cn):
    try:
        c = cn.cursor()
        c.execute("""
            select version
            from migrations
            order by id desc
            limit 1
        """)
        return c.fetchone()[0]
    except Exception:
        return 0
